from collections import namedtuple, OrderedDict
import re
from datetime import datetime
from wq.io.exceptions import NoData, MappingFailed


class BaseMapper(object):
    def get_key_field(self):
        return self.map_field(self.key_field)

    def map_field(self, field):
        return field

    def map_value(self, field, value):
        return value

    def unmap_field(self, field):
        return field

    def unmap_value(self, field, value):
        return value

    def usable_item(self, item):
        uitem = {}
        for key, val in item.items():
            field = self.map_field(key)
            value = self.map_value(field, val)
            uitem[field] = value
        return uitem

    def parse_usable_item(self, uitem):
        item = {}
        for field, value in uitem.items():
            key = self.unmap_field(field)
            val = self.unmap_value(field, value)
            item[key] = val
        return item


class DictMapper(BaseMapper):
    field_map = {}
    value_map = {}

    def map_field(self, field):
        field = self.field_map[field] if field in self.field_map else field
        return field

    def map_value(self, field, value):
        if not isinstance(value, str):
            return value
        value = self.value_map[value] if value in self.value_map else value
        return value

    def unmap_field(self, field):
        for f in self.field_map:
            if self.field_map[f] == field:
                return f
        return field

    def unmap_value(self, field, value):
        if not isinstance(value, str):
            return value
        for v in self.value_map:
            if self.value_map[v] == value:
                return v
        return value


class TupleMapper(DictMapper):
    no_pickle_mapper = ['_tuple_class', '_tuple_prototype']

    @property
    def field_map(self):
        field_names = self.get_field_names()
        if not field_names and not getattr(self, 'data', None):
            raise NoData

        # FIXME: check for duplicates
        if not hasattr(self, '_field_map'):
            items = [
                (field, self.tuple_field_name(field))
                for field in field_names
            ]
            self._field_map = OrderedDict(items)
        return self._field_map

    def tuple_field_name(self, field):
        field = self.clean_field_name(field)
        return re.sub(r'\W', '', field.lower())

    def clean_field_name(self, field):
        return field

    @property
    def tuple_class(self):
        "Returns a class to use for individual items"

        if not hasattr(self, '_tuple_class'):
            cls = namedtuple(
                self.__class__.__name__ + 'Tuple',
                list(self.field_map.values())
            )
            self._tuple_class = cls

        return self._tuple_class

    @property
    def tuple_prototype(self):
        if not hasattr(self, '_tuple_prototype'):
            vals = {field: None for field in self.field_map.values()}
            self._tuple_prototype = self.tuple_class(**vals)
        return self._tuple_prototype

    def usable_item(self, item):
        mapped = super(TupleMapper, self).usable_item(item)
        try:
            return self.tuple_prototype._replace(**mapped)
        except ValueError as e:
            raise MappingFailed(str(e))

    def parse_usable_item(self, uitem):
        mapped = {key: getattr(uitem, key) for key in self.field_map.values()}
        return super(TupleMapper, self).parse_usable_item(mapped)

    def item_dict(self, uitem):
        return uitem._asdict()

    def create(self, **kwargs):
        return self.tuple_prototype._replace(**kwargs)


def make_date_mapper(fmt):
    """
    Generate functions to use for mapping strings to dates
    """
    def mapper(val):
        val = datetime.strptime(val, fmt)
        if '%Y' in fmt or '%y' in fmt:
            return val
        else:
            return val.time()
    return mapper


class TimeSeriesMapper(TupleMapper):
    date_formats = None
    map_floats = True

    def map_value(self, field, value):
        if not isinstance(value, str):
            return value

        functions = [make_date_mapper(fmt) for fmt in self.date_formats]
        if self.map_floats:
            functions.insert(0, float)

        value = value.strip()
        for i, fn in enumerate(functions):
            try:
                return fn(value)
            except ValueError:
                pass
        return value

    @property
    def key_fields(self):
        raise NotImplementedError("Key fields must be specified")

    def parameter_fields(self):
        return sorted(set(self.field_map.values()) - set(self.key_fields))



from wq.io.base import BaseIO
from wq.io.loaders import FileLoader, BinaryFileLoader, NetLoader, StringLoader
from wq.io.parsers import CsvParser, JsonParser, XmlParser, ExcelParser
from wq.io.mappers import TupleMapper
import mimetypes

PARSERS = {
    'application/vnd.ms-excel': ExcelParser,
    'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet':
    ExcelParser,
    'text/csv': CsvParser,
    'application/json': JsonParser,
    'application/xml': XmlParser,
}
BINARY_FORMATS = (
    'application/xml',
    'application/vnd.ms-excel',
    'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
)


# Save generated classes to avoid recreating them
_io_classes = {}


def make_io(loader, parser, mapper=TupleMapper, name=None, module="wq.io"):
    """
    Mix the specified loader, parser, and mapper classes into a usable IO
    """
    key = (loader, parser, mapper)
    if key in _io_classes:
        return _io_classes[key]

    if name is None:
        lname = parser.__name__.replace('Parser', '')
        pname = loader.__name__.replace('Loader', '')
        if mapper == TupleMapper:
            mname = ""
        else:
            mname = mapper.__name__.replace('Mapper', '')
        name = lname + pname + mname + "IO"
    cls = type(name, (loader, parser, mapper, BaseIO), {})
    cls.__module__ = module
    _io_classes[key] = cls
    return cls


def guess_type(filename, buffer=None):
    mimetype, encoding = mimetypes.guess_type(filename)
    if mimetype is None:
        try:
            import magic
            if buffer:
                mimetype = magic.from_buffer(buffer, mime=True)
            else:
                mimetype = magic.from_file(filename, mime=True)
        except ImportError:
            pass
    return mimetype


def load_file(filename, mapper=TupleMapper, options={}):
    mimetype = guess_type(filename)
    if mimetype not in PARSERS:
        raise Exception("Could not determine parser for %s" % mimetype)
    parser = PARSERS[mimetype]
    loader = BinaryFileLoader if mimetype in BINARY_FORMATS else FileLoader
    IO = make_io(loader, parser, mapper)
    return IO(filename=filename, **options)


def load_string(string, mapper=TupleMapper, options={}):
    if string.startswith('<'):
        parser = XmlParser
    elif string.startswith('[') or (
            string.startswith('{') and 'namespace' in options):
        parser = JsonParser
    elif ',' in string:
        parser = CsvParser
    else:
        raise Exception("Could not determine parser for string!")

    loader = StringLoader
    IO = make_io(loader, parser, mapper)
    return IO(string=string, **options)


class FlatIO(TupleMapper, BaseIO):
    """
    Denormalizes a nested IO structure (e.g. an array of individual time
    series) into a single iterable.  Each row in the top level IO should have
    an attribute (typically 'data') pointing to an inner IO.  Both the top
    level IO class and the inner class should extend TupleMapper.
    """

    io_class = None
    inner_attr = 'data'

    def __init__(self, *args, **kwargs):
        self.io_class = kwargs.pop('io_class', self.io_class)
        self.inner_attr = kwargs.pop('inner_attr', self.inner_attr)
        if self.io_class is None:
            raise Exception("An IO class must be specified")

        # Pass remaining arguments
        self.nested_io = self.io_class(*args, **kwargs)
        self.data = list(self.unpack_io())

    def unpack_io(self):
        # Loop through outer IO (e.g. metadata series)
        for outer in self.nested_io:
            meta = outer._asdict()
            inner_io = meta.pop(self.inner_attr)

            # Loop through inner IO (e.g. time series) on each record
            for inner in inner_io:
                record = meta.copy()
                record.update(inner._asdict())
                yield record


def flattened(io_class, *args, **kwargs):
    if io_class.nested:
        return FlatIO(io_class=io_class, *args, **kwargs)
    else:
        return io_class(*args, **kwargs)



try:
    from bs4 import BeautifulSoup
except ImportError:
    BeautifulSoup = None


class IoException(Exception):
    def __str__(self):
        if self.args and self.args[0]:
            return self.args[0]
        return self.__doc__


class LoadFailed(IoException):
    """Error loading data!"""

    def __init__(self, message, path=None, code=None):
        super(LoadFailed, self).__init__(message)
        self.path = path
        self.code = code

    def __str__(self):
        if self.args and self.args[0]:
            text = self.args[0]
            has_html = False
            for tag in '<html', '<body', '<div':
                if tag in text or tag.upper() in text:
                    has_html = True
            if has_html and BeautifulSoup:
                html = BeautifulSoup(text).body
                if html:
                    text = html.get_text('\n')
            return text
        elif self.code is not None:
            return "%s Error" % self.code
        return super(LoadFailed, self).__str__()


class ParseFailed(IoException):
    """Error parsing data!"""
    pass


class MappingFailed(IoException):
    """Error processing data!"""
    pass


class NoData(IoException):
    """No data returned!"""
    pass



import fiona
from shapely import wkt, geometry
from wq.io.loaders import FileLoader
from wq.io.parsers.base import BaseParser
from wq.io.mappers import TupleMapper


class FionaLoaderParser(FileLoader, BaseParser):
    """
    Composite loader & parser mixin for GIS data, powered by Fiona
    """
    layer_id = None
    meta = {}
    key_field = 'id'

    def load(self):
        try:
            self.layers = fiona.listlayers(self.filename)
        except (ValueError, IOError) as e:
            driver = guess_driver(self.filename)
            self.meta = {'driver': driver}
            self.empty_file = True

    def parse(self):
        # If multiple layers, parse all of them (!)
        if len(self.layers) > 1 and self.layer_id is None:
            cls = type(self)
            self.data = [{
                'id': id,
                'name': name,
                'data': cls(filename=self.filename, layer_id=id)
            } for id, name in enumerate(self.layers)]
        else:
            # One layer, load & parse GIS data
            with fiona.open(self.filename, layer=self.layer_id) as f:
                self.meta = f.meta
                self.data = list(map(self.parse_feature, f))

    def parse_feature(self, f):
        # Flatten Fiona's GeoJSON-style representation into something more
        # amenable to namedtuple-ing
        feat = {key: value for key, value in f['properties'].items()}
        if 'id' not in feat and 'ID' not in feat:
            feat['id'] = f['id']
        feat['geometry'] = f['geometry']
        return feat

    def dump_feature(self, feat, i):
        # Undo aforementioned flattening
        return {
            'id': feat.get('id', feat.get('ID', i)),
            'geometry': feat['geometry'],
            'properties': {
                key: value for key, value in feat.items()
                if key not in ('geometry', 'id',)
            }
        }

    def dump(self):
        # Dump and save the dataset at the same time via Fiona
        pass

    def save(self):
        with fiona.open(self.filename, 'w', **self.meta) as f:
            for i, feat in enumerate(self.data):
                f.write(self.dump_feature(feat, i))


class GisMapper(TupleMapper):
    """
    GIS-aware tuple mapper
    """
    def as_dataframe(self):
        # Mimic BaseIO.as_dataframe() but with GeoDataFrame
        # (also, key_field is always set)
        from geopandas import GeoDataFrame
        key = self.get_key_field()
        data = [self.item_dict(row) for row in self.values()]
        df = GeoDataFrame(data)
        df.set_index(key, inplace=True)
        return df

    def item_dict(self, uitem):
        # Turn usable item into GeoDataFrame-friendly dict
        data = uitem._asdict()
        data['geometry'] = geometry.shape(data['geometry'])
        return data


class ShapeMapper(GisMapper):
    """
    Map Fiona's GeoJSON-style geometries to and from Shapely shapes
    """
    def map_value(self, field, value):
        value = super(ShapeMapper, self).map_value(field, value)
        if field == 'geometry':
            value = geometry.shape(value)
        return value

    def unmap_value(self, field, value):
        if field == 'geometry':
            value = geometry.mapping(value)
        return super(ShapeMapper, self).unmap_value(field, value)

    def item_dict(self, uitem):
        return uitem._asdict()


class WktMapper(ShapeMapper):
    """
    Map geometries to and from WKT (good for Django integration)
    """
    def map_value(self, field, value):
        value = super(WktMapper, self).map_value(field, value)
        if field == 'geometry':
            value = wkt.dumps(value)
        return value

    def unmap_value(self, field, value):
        if field == 'geometry':
            value = wkt.loads(value)
        return super(WktMapper, self).unmap_value(field, value)

    def item_dict(self, uitem):
        data = uitem._asdict()
        data['geometry'] = wkt.loads(data['geometry'])
        return data


def guess_driver(filename):
    if filename.endswith(".shp"):
        return "ESRI Shapefile"
    else:
        return "GeoJSON"



from wq.io.base import *
from wq.io.loaders import *
from wq.io.parsers import *
from wq.io.mappers import *
from wq.io.util import make_io, load_file, load_string, guess_type
from wq.io.version import VERSION

# Some useful pre-mixed classes
CsvFileIO = make_io(FileLoader, CsvParser)
CsvNetIO = make_io(NetLoader, CsvParser)
CsvStringIO = make_io(StringLoader, CsvParser)

JsonFileIO = make_io(FileLoader, JsonParser)
JsonNetIO = make_io(NetLoader, JsonParser)
JsonStringIO = make_io(StringLoader, JsonParser)

XmlFileIO = make_io(BinaryFileLoader, XmlParser, name='XmlFileIO')
XmlNetIO = make_io(NetLoader, XmlParser)
XmlStringIO = make_io(StringLoader, XmlParser)

ExcelFileIO = make_io(BinaryFileLoader, ExcelParser, name='ExcelFileIO')

try:
    import fiona
    import shapely
    from wq.io.gis import GisIO, ShapeIO, WktIO
except ImportError:
    pass



import json
from .readers import csv, UNICODE_CSV, SkipPreludeReader
from xml.etree import ElementTree as ET

from .base import BaseParser, TableParser
from wq.io.exceptions import ParseFailed


class CsvParser(TableParser):
    delimiter = ","
    quotechar = '"'
    no_pickle_parser = ['csvdata']

    def parse(self):
        # Like DictReader, assume explicit field definition means CSV does not
        # contain column headers.
        fields = self.get_field_names()
        if self.start_row is None:
            if fields:
                self.start_row = 0
            else:
                self.start_row = 1
        if self.header_row is None:
            if fields:
                self.header_row = None
            else:
                self.header_row = 0

        Reader = self.reader_class()
        self.csvdata = Reader(
            self.file,
            fields,
            delimiter=self.delimiter,
            quotechar=self.quotechar,
        )
        self.field_names = self.csvdata.fieldnames
        if self.header_row is not None:
            self.header_row = self.csvdata.header_row
        self.data = [row for row in self.csvdata]

    def reader_class(self):
        class Reader(SkipPreludeReader):
            max_header_row = self.max_header_row
        return Reader

    def dump(self, file=None):
        if file is None:
            file = self.file
        args = file, self.get_field_names()
        kwargs = {'encoding': 'utf-8'} if UNICODE_CSV else {}
        kwargs['delimiter'] = self.delimiter
        kwargs['quotechar'] = self.quotechar
        csvout = csv.DictWriter(*args, **kwargs)
        csvout.writeheader()
        for row in self.data:
            csvout.writerow(row)


class JsonParser(BaseParser):
    indent = None
    namespace = None

    def parse(self):
        try:
            obj = json.load(self.file)
            if self.namespace:
                for key in self.namespace.split('.'):
                    obj = obj[key]
            self.data = list(map(self.parse_item, obj))
        except ValueError:
            raise ParseFailed

    def parse_item(self, item):
        return item

    def dump(self, file=None):
        if file is None:
            file = self.file
        obj = list(map(self.dump_item, self.data))
        if self.namespace:
            for key in reversed(self.namespace.split('.')):
                obj = {key: obj}
        json.dump(obj, file, indent=self.indent)

    def dump_item(self, item):
        return item


class XmlParser(BaseParser):
    root_tag = None
    item_tag = None

    def parse(self):
        doc = ET.parse(self.file)
        root = self.parse_root(doc)
        if self.root_tag is None:
            self.root_tag = root.tag
        if self.item_tag is None:
            self.item_tag = list(root)[0].tag
        self.data = list(map(self.parse_item, root.findall(self.item_tag)))

    def parse_root(self, doc):
        root = doc.getroot()
        if self.root_tag is not None and root.tag != self.root_tag:
            root = root.find(self.root_tag)
        return root

    def parse_item(self, el):
        return {e.tag: e.text for e in el}

    def dump(self, file=None):
        if file is None:
            file = self.file
        root = ET.Element(self.root_tag)
        for item in self.data:
            root.append(self.dump_item(item))
        ET.ElementTree(root).write(file)

    def dump_item(self, item):
        el = ET.Element(self.item_tag)
        for key in self.get_field_names():
            if key not in item or item[key] is None:
                continue
            sel = ET.SubElement(el, key)
            sel.text = str(item.get(key))
        return el



from .mixins import FionaLoaderParser, GisMapper, ShapeMapper, WktMapper
from wq.io.base import BaseIO


class MetaSyncIO(BaseIO):
    """
    Custom sync() to handle transfering Fiona metadata (except for driver)
    """
    def sync(self, other, save=True):
        driver = other.meta.get('driver', None)
        other.meta = self.meta.copy()
        if driver:
            other.meta['driver'] = driver
        super(MetaSyncIO, self).sync(other, save)

    def get_field_names(self):
        if self.field_names is None and self.meta is not None:
            return (
                ['id', 'geometry']
                + list(self.meta['schema']['properties'].keys())
            )
        return super(MetaSyncIO, self).get_field_names()


class GisIO(FionaLoaderParser, GisMapper, MetaSyncIO):
    pass


class ShapeIO(FionaLoaderParser, ShapeMapper, MetaSyncIO):
    pass


class WktIO(FionaLoaderParser, WktMapper, MetaSyncIO):
    pass



from collections import MutableMapping, MutableSequence


class BaseIO(MutableMapping, MutableSequence):
    "wq.io.BaseIO: Base class for generic resource management"

    tabular = False
    nested = False

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)
        self.refresh()

    def refresh(self):
        self.load()
        if getattr(self, 'empty_file', False):
            self.data = []
        else:
            self.parse()
            if hasattr(self, 'file') and not self.file.closed:
                self.file.close()

    def load(self):
        "Open a resource (defined by loader mixins)"
        # self.file = ...
        pass

    def parse(self):
        """
        Parse a resource (defined by parser mixins).
        Result should be an iterable of dicts.
        """
        # self.data = some_parse_method(self.file)
        pass

    def dump(self, file=None):
        ""
        if file is None:
            file = self.file
        file.write(str(self.data))

    def save(self):
        ""
        self.dump(self.file)

    field_names = None
    scan_fields = False
    _auto_field_names = None

    def get_field_names(self):
        "Returns a list of raw fields to expect (defined by parser mixins)"
        if self.field_names is not None:
            # Support specifying field_names as string (like namedtuple does)
            if isinstance(self.field_names, str):
                return self.field_names.replace(',', ' ').split()
            else:
                return self.field_names

        # If no defined field names, try to retrieve from data
        if not getattr(self, 'data', None):
            return None

        if self._auto_field_names:
            return self._auto_field_names

        if self.scan_fields:
            # Scan all rows for field names
            field_names = set()
            for row in self.data:
                field_names.update(row.keys())
            field_names = list(field_names)
        else:
            # Assume first row contains same keys as all other rows
            field_names = list(self.data[0].keys())

        self._auto_field_names = field_names
        return field_names

    @property
    def key_field(self):
        "Assign a key_field to use the resource as a Map"
        return None

    def get_key_field(self):
        return self.key_field

    def usable_item(self, item):
        "Hook to allow items to be transformed"
        return item

    def parse_usable_item(self, uitem):
        "Hook to allow items to be untransformed"
        return uitem

    def compute_index(self, recompute=False):
        key_field = self.get_key_field()
        if key_field is None:
            return None

        if getattr(self, '_index_cache', None) is not None and not recompute:
            return self._index_cache

        index = {}
        for i, item in enumerate(self.data):
            uitem = self.usable_item(item)
            if isinstance(uitem, dict):
                key = uitem.get(key_field, None)
            else:
                key = getattr(uitem, key_field, None)
            if key is not None:
                index[key] = i

        self._index_cache = index
        return index

    def find_index(self, key):
        index = self.compute_index()
        if index is not None:
            return index.get(key, None)
        else:
            return key

    def __len__(self):
        return len(self.data)

    def __getitem__(self, key):
        index = self.find_index(key)
        if index is None:
            raise KeyError
        return self.usable_item(self.data[index])

    def __setitem__(self, key, uitem):
        item = self.parse_usable_item(uitem)
        index = self.find_index(key)
        if index is not None:
            self.data[index] = item
        else:
            self.data.append(item)
            self.compute_index(True)

    def __delitem__(self, key):
        index = self.find_index(key)
        if index is None:
            raise KeyError
        del self.data[index]
        self.compute_index(True)

    def insert(self, index, uitem):
        item = self.parse_usable_item(uitem)
        self.data.insert(index, item)
        self.compute_index(True)

    def __iter__(self):
        for item in self.data:
            uitem = self.usable_item(item)
            if uitem is None:
                return
            pk = self.get_key_field()
            if pk is None:
                yield uitem
            elif isinstance(uitem, dict):
                yield uitem.get(pk, None)
            else:
                yield getattr(uitem, pk, None)

    def sync(self, other, save=True):
        if self.get_key_field() is None or other.get_key_field() is None:
            raise Exception("Key field required to sync!")
        for key in self:
            other[key] = self[key]
        if save:
            other.save()

    def copy(self, other, save=True):
        del other.data[:]
        for item in self.data:
            uitem = self.usable_item(item)
            other.append(uitem)
        if save:
            other.save()

    # Slots to track things that can't be pickled
    # (need a separate slot for each expected mixin class, since they don't
    #  extend BaseIO)
    no_pickle = []
    no_pickle_loader = []
    no_pickle_mapper = []
    no_pickle_parser = []

    def get_no_pickle(self):
        return (
            self.no_pickle +
            self.no_pickle_loader +
            self.no_pickle_mapper +
            self.no_pickle_parser
        )

    def __getstate__(self):
        """
        Don't include auto-created and unpicklable properties in state.
        """
        state = self.__dict__.copy()
        for name in self.get_no_pickle():
            state.pop(name, None)
        return state

    def item_dict(self, item):
        return item

    def as_dataframe(self):
        from pandas import DataFrame
        key = self.get_key_field()
        if key:
            data = [self.item_dict(row) for row in self.values()]
        else:
            data = [self.item_dict(row) for row in self]
        df = DataFrame(data)
        if key:
            df.set_index(key, inplace=True)
        return df



from __future__ import print_function
import requests
try:
    # Python 2 (uses str)
    from StringIO import StringIO
except ImportError:
    # Python 3 (Python 2 equivalent uses unicode)
    from io import StringIO
from io import BytesIO
from wq.io.version import VERSION
from .exceptions import LoadFailed
from zipfile import ZipFile


class BaseLoader(object):
    no_pickle_loader = ['file']
    empty_file = None
    binary = False

    def load(self):
        raise NotImplementedError


class FileLoader(BaseLoader):
    filename = None
    read_mode = 'r'
    write_mode = 'w+'

    def load(self):
        try:
            self.file = open(self.filename, self.read_mode)
            self.empty_file = False
        except IOError:
            if self.binary:
                self.file = BytesIO()
            else:
                self.file = StringIO()
            self.empty_file = True

    def save(self):
        file = open(self.filename, self.write_mode)
        self.dump(file)
        file.close()


class BinaryFileLoader(FileLoader):
    binary = True
    read_mode = 'rb'
    write_mode = 'wb+'


class Zipper(object):
    inner_filename = None
    inner_binary = False

    def unzip_file(self):
        zipfile = ZipFile(self.file)
        inner_file = zipfile.read(
            self.get_inner_filename(zipfile)
        )
        if self.inner_binary:
            self.file = BytesIO(inner_file)
        else:
            self.file = StringIO(inner_file.decode('utf-8'))
        zipfile.fp.close()
        zipfile.close()

    def get_inner_filename(self, zipfile):
        if self.inner_filename:
            return self.inner_filename
        names = zipfile.namelist()
        if len(names) == 1:
            return names[0]

        zipfile.fp.close()
        zipfile.close()
        raise LoadFailed("Multiple Inner Files!")


class ZipFileLoader(Zipper, BinaryFileLoader):
    def load(self):
        super(ZipFileLoader, self).load()
        self.unzip_file()


class StringLoader(BaseLoader):
    string = ""

    def load(self):
        self.file = StringIO(self.string)

    def save(self):
        file = StringIO()
        self.dump(file)
        self.string = file.getvalue()
        file.close()


class NetLoader(BaseLoader):
    "NetLoader: opens HTTP/REST resources for use in wq.io"

    username = None
    password = None
    debug = False

    @property
    def url(self):
        raise NotImplementedError

    @property
    def user_agent(self):
        return "wq.io/%s (%s)" % (VERSION, requests.utils.default_user_agent())

    @property
    def headers(self):
        return {
            'User-Agent': self.user_agent,
        }

    def load(self, **kwargs):
        result = self.GET()
        if self.binary:
            self.file = BytesIO(result)
        else:
            self.file = StringIO(result)

    def req(self, url=None, method=None, params=None, body=None, headers={}):
        if url is None:
            url = self.url

        if params is None:
            params = getattr(self, 'params', None)

        if isinstance(params, str):
            url += '?' + params
            params = None

        if self.debug:
            if params:
                from requests.compat import urlencode
                debug_url = url + '?' + urlencode(params, doseq=True)
            else:
                debug_url = url
            self.debug_string = "%s: %s" % (method, debug_url)
            print(self.debug_string)

        if self.username is not None and self.password is not None:
            auth = (self.username, self.password)
        else:
            auth = None

        all_headers = self.headers.copy()
        all_headers.update(headers)

        resp = requests.request(
            method, url,
            params=params,
            headers=all_headers,
            auth=auth,
            data=body,
        )

        if resp.status_code < 200 or resp.status_code > 299:
            raise LoadFailed(
                resp.text,
                path=url,
                code=resp.status_code,
            )

        if self.binary:
            return resp.content
        else:
            return resp.text

    def GET(self, **kwargs):
        return self.req(method='GET', **kwargs)

    def POST(self, **kwargs):
        return self.req(method='POST', **kwargs)

    def PUT(self, **kwargs):
        return self.req(method='PUT', **kwargs)

    def DELETE(self, **kwargs):
        return self.req(method='DELETE', **kwargs)


class ZipNetLoader(Zipper, NetLoader):
    binary = True

    def load(self):
        super(ZipNetLoader, self).load()
        self.unzip_file()



try:
    import unicodecsv as csv
    UNICODE_CSV = True
except ImportError:
    import csv
    UNICODE_CSV = False


if issubclass(csv.DictReader, object):
    # Python 3
    DictReader = csv.DictReader
else:
    # Python 2
    class DictReader(object, csv.DictReader):
        pass


class SkipPreludeReader(DictReader):
    """
    A specialized version of DictReader that attempts to find where the "real"
    CSV data is in a file that may contain a prelude of non-CSV text.
    """

    max_header_row = 20

    def __init__(self, f, fieldnames=None, restkey=None, restval=None,
                 dialect="excel", *args, **kwds):
        # Preserve file since we're going to start reading it
        self._file = f

        # Preserve reader options since we'll need to make another one
        readeropts = [f, dialect]
        readeropts.extend(args)
        self._readeropts = (readeropts, kwds)
        csv.DictReader.__init__(self, f, fieldnames, restkey, restval,
                                dialect, *args, **kwds)

    @property
    def fieldnames(self):
        if self._fieldnames is not None:
            return self._fieldnames

        # Create a new reader just to figure out which row is the header
        args, kwds = self._readeropts
        data = csv.reader(*args, **kwds)
        rows = []
        for i in range(self.max_header_row):
            try:
                rows.append(next(data))
            except StopIteration:
                pass
        header_row, field_names = self.choose_header(rows)

        # Reset file and advance reader so it starts in the right spot
        self._file.seek(0)
        for i in range(header_row + 1):
            try:
                next(self.reader)
            except StopIteration:
                pass

        self._fieldnames = field_names
        self._header_row = header_row
        return field_names

    @property
    def header_row(self):
        self.fieldnames  # used for side effect
        return self._header_row

    def choose_header(self, rows):
        """
        Determine which row contains column headers from the provided set.
        Default is to assume that the first longest row is the header.
        """
        header_row = 0
        field_names = []

        # Select header from available rows
        for i, row in enumerate(rows):
            if len(row) > len(field_names):
                header_row = i
                field_names = row
        return header_row, field_names



import xlrd
import datetime
import math
from .base import TableParser


class WorkbookParser(TableParser):
    workbook = None
    worksheet = None
    sheet_name = 0
    start_row = None
    column_count = None
    no_pickle_parser = ['workbook', 'worksheet']

    def parse(self):
        self.parse_workbook()
        if self.sheet_name is None:
            self.data = [{'name': name, 'data': self.get_sheet_by_name(name)}
                         for name in self.sheet_names]
            return

        sheet_name = self.sheet_name
        if isinstance(self.sheet_name, int):
            sheet_name = self.sheet_names[sheet_name]

        self.parse_worksheet(sheet_name)

        if self.header_row is None:
            if self.start_row is not None:
                self.header_row = self.start_row - 1
            else:
                self.column_count = 0

                def checkval(cell):
                    if cell.value is not None and cell.value != '':
                        return True
                    return False

                search_rows = min(len(self.worksheet) - 1, self.max_header_row)
                for row in range(search_rows, -1, -1):
                    count = len(list(filter(checkval, self.worksheet[row])))
                    if count >= self.column_count:
                        self.column_count = count
                        self.header_row = row

        if self.start_row is None:
            self.start_row = self.header_row + 1

        if self.field_names is None:
            rows = self.worksheet[self.header_row:self.start_row]
            self.field_names = [
                str(c.value) or 'c%s' % i for i, c in enumerate(rows[0])
            ]
            for row in rows[1:]:
                for i, c in enumerate(row):
                    self.field_names[i] += "\n" + str(c.value)

            seen_fields = set()
            for i, field in enumerate(self.field_names):
                if field in seen_fields:
                    field += str(i)
                    self.field_names[i] = field
                seen_fields.add(field)

        self.data = list(map(self.parse_row, self.worksheet[self.start_row:]))
        if self.header_row > 0:
            for r in range(0, self.header_row):
                for c, cell in enumerate(self.worksheet[r]):
                    val = self.get_value(cell)
                    if val is not None and val != '':
                        self.extra_data.setdefault(r, {})
                        self.extra_data[r][c] = val

    def parse_workbook(self):
        raise NotImplementedError

    @property
    def sheet_names(self):
        raise NotImplementedError

    def get_sheet_by_name(self, name):
        raise NotImplementedError

    def parse_worksheet(self, name):
        raise NotImplementedError

    def parse_row(self, row):
        raise NotImplementedError

    def get_value(self, cell):
        raise NotImplementedError

    def dump(self, file=None):
        if file is None:
            file = self.file
        write, close = self.open_worksheet(file)
        for i, field in enumerate(self.field_names):
            write(0, i, field)
        for r, row in enumerate(self.data):
            for c, field in enumerate(self.field_names):
                write(r + 1, c, row[field])
        close()


class ExcelParser(WorkbookParser):
    def parse_workbook(self):
        self.workbook = xlrd.open_workbook(file_contents=self.file.read())

    @property
    def sheet_names(self):
        return self.workbook.sheet_names()

    def get_sheet_by_name(self, name):
        return self.workbook.sheet_by_name(name)

    def parse_worksheet(self, name):
        worksheet = self.get_sheet_by_name(name)
        self.worksheet = [worksheet.row(i) for i in range(worksheet.nrows)]

    def parse_row(self, row):
        return {name: self.get_value(row[i])
                for i, name in enumerate(self.get_field_names())
                if i < len(row)}

    def get_value(self, cell):
        if cell.ctype == xlrd.XL_CELL_DATE:
            time, date = math.modf(cell.value)
            tpl = xlrd.xldate_as_tuple(cell.value, self.workbook.datemode)
            if date and time:
                return datetime.datetime(*tpl)
            elif date:
                return datetime.date(*tpl[0:3])
            else:
                return datetime.time(*tpl[3:6])
        return cell.value

    def open_worksheet(self, file):
        if getattr(self, 'filename', '').endswith('.xls'):
            import xlwt
            workbook = xlwt.Workbook()
            worksheet = workbook.add_sheet('Sheet 1')

            def close():
                workbook.save(file)
        else:
            import xlsxwriter
            workbook = xlsxwriter.Workbook(file)
            worksheet = workbook.add_worksheet()

            def close():
                workbook.close()
        return worksheet.write, close
